/*
 * Transformations on STLC with recursion
 * Date Created:   October 15, 2014
*/

sig trans.

kind nat type.
type z nat.
type s nat -> nat.

type add   nat -> nat -> nat -> o.
type npred  nat -> nat -> o.
type is_nat  nat -> o.

% Source and target types are in the same category.
kind ty type.

type tnat    ty.                 % for natural numbers
type arr     ty -> ty -> ty.     % for abstractions in source and
                                 %% closures in the target languages.
type tunit   ty.                 % for units
type prod    ty -> ty -> ty.     % for pairs 
type arr'    ty -> ty -> ty.     % for abstractions in the target language


% Source Terms
kind tm type.

type nat         nat -> tm.
type pred        tm -> tm.
type plus        tm -> tm -> tm.
type ifz         tm -> tm -> tm -> tm.
type unit        tm.
type pair        tm -> tm -> tm.
type fst, snd    tm -> tm.
type let         tm -> (tm -> tm) -> tm.
type fix         (tm -> tm -> tm) -> tm.
type app         tm -> tm -> tm.

type bnd         (tm -> tm) -> tm.
type base        tm -> tm.

% Target Terms
kind tm' type.
    
type nat'         nat -> tm'.
type pred'        tm' -> tm'.
type plus'        tm' -> tm' -> tm'.
type ifz'         tm' -> tm' -> tm' -> tm'.
type unit'        tm'.
type pair'        tm' -> tm' -> tm'.
type fst', snd'   tm' -> tm'.
type let'         tm' -> (tm' -> tm') -> tm'.
type abs'         (tm' -> tm') -> tm'.
type app'         tm' -> tm' -> tm'.
type clos'        tm' -> tm' -> tm'.
type open'        tm' -> (tm' -> tm' -> tm') -> tm'.

% CMinor like code
type dummyval     tm'.
type loc'         nat -> tm'.
type alloc'       nat -> tm'.
type move'        tm' -> nat -> tm' -> tm'.
type load'        tm' -> nat -> tm'.

% Terms
type tm          tm -> o.
type tm'         tm' -> o.
type tm''        tm' -> o.
type tm''_body   tm' -> o.

% List
type memb    A -> list A -> o.
type appd    list A -> list A -> list A -> o.
type combine list A -> list A -> list A -> o.
type rev list A -> list A -> list A -> o.

% Mapping 
kind map type -> type -> type.
type map A -> B -> map A B.

% Values
type val         tm -> o.
type val'        tm' -> o.
type tm'_list_val   list tm' -> o.
type val''       tm' -> o.

% Typing rules
type of          tm -> ty -> o.
type ofc'        tm' -> ty -> o.
type of'         tm' -> ty -> o.
type of''        tm' -> ty -> o.
type of''_env    list tm' -> ty -> o.
type of''_body   ty -> tm' -> ty -> o.

% Well-formed types
type is_sty       ty -> o.
type is_cty       ty -> o.

% Call-by-value evaluation rules in small steps
type step        tm -> tm -> o.
type nstep       nat -> tm -> tm -> o.
type eval        tm -> tm -> o.

type step'        tm' -> tm' -> o.
type nstep'       nat -> tm' -> tm' -> o.
type eval'        tm' -> tm' -> o.
type eval''       tm' -> tm' -> o.

% Evaluation semantics of generated code using heaps
kind state        type.
type st           nat -> (list (map nat tm')) -> state.
type allocate     nat -> nat -> (list (map nat tm')) -> (list (map nat tm')) -> o.
type update_heap  (list (map nat tm')) -> nat -> tm' -> (list (map nat tm')) -> o.
type lookup_heap  (list (map nat tm')) -> nat -> tm' -> o.
type step''       state -> tm' -> state -> tm' -> o.
type nstep''      nat -> state -> tm' -> state -> tm' -> o.
type heval        tm' -> state -> tm' -> o.
type eval'''      tm' -> state -> tm' -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          CPS Transformation                 %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
type cps  tm -> (tm -> tm) -> tm -> o.
type cps' tm -> tm -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          Closure Conversion                 %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Projection from free variables to an environment parameter
type mapvar list tm -> (tm' -> list (map tm tm')) -> o.

% Forming the closure environment
type mapenv list tm -> list (map tm tm') -> tm' -> o.

% Collecting free variables in an abstraction
type notfree tm -> o.
type fvars tm -> list tm -> list tm -> o.

% Closure conversion
% The first argument is a mapping from free variables to target terms 
% and the second argument is a list of these free variables
type cc    list (map tm tm') -> list tm -> tm -> tm' -> o.
type cc'   tm -> tm' -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          Code Hoisting                      %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Hoisted terms consisting of a list of closed functions and the body
type hbase     tm' -> tm'.             % Base term of the body
type habs      (tm' -> tm') -> tm'.    % Top-level bindings of the body
type htm       list tm' -> tm' -> tm'. % The hoisted term

% Combining the bodies of two hoisted terms
type hcombine      tm' -> tm' -> (tm' -> tm' -> tm') -> tm' -> o.
type hcombine3     tm' -> tm' -> tm' ->
                   (tm' -> tm' -> tm' -> tm') -> tm' -> o.
type hcombine_abs  tm' -> (tm' -> tm') -> 
                   (tm' -> (tm' -> tm') -> tm') -> tm' -> o.

% Apply the construct under the abstractions of 
% the body hoisted terms.
type hconstr       tm' -> (tm' -> tm') -> tm' -> o.

% Extract functions out of an abstraction and produce
% a closed abstraction
type hoist_abs     (tm' -> tm') -> tm' -> tm' -> list tm' -> 
                   tm' -> (tm' -> tm') -> o.
type abstract   (tm' -> tm') -> (tm' -> tm') -> (tm' -> tm' -> tm') -> o.

type tm'_list_to_tuple     list tm' -> tm' -> o.

% Code hoisting
type ch        tm' -> tm' -> o.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          Code Generation                    %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

type cgen       tm' -> (tm' -> tm') -> tm' -> o.

type ecgen      list tm' -> list tm' -> o.
type bcgen      tm' -> tm' -> o.
type hcgen'     tm' -> tm' -> o.
type hcgen      tm' -> tm' -> o.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%          Full Compilation                   %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

type compile    tm -> tm' -> o.